# coding: utf-8

# $Id: $
import re


def sphinx_escape(value):
    """ Escapes SphinxQL search expressions. """

    if not isinstance(value, str):
        return value

    value = re.sub(r"([=<>()|!@~&/^$\-\'\"\\])", r'\\\\\\\1', value)
    value = re.sub(r'\b(SENTENCE|PARAGRAPH)\b', r'\\\\\\\1', value)
    return value



# coding: utf-8
import re

from django.core.exceptions import FieldError
from django.db import models
from django.db.models.expressions import Random
from django.db.models.lookups import Exact
from django.db.models.sql import compiler, AND
from django.db.models.sql.constants import ORDER_DIR
# noinspection PyProtectedMember
from django.db.models.sql.query import get_order_dir

from sphinxsearch import sql as sqls


class SphinxQLCompiler(compiler.SQLCompiler):
    # Options names that are not escaped by compiler. Don't pass user input
    # there.
    safe_options = ('ranker', 'field_weights', 'index_weights')

    def compile(self, node, select_format=False):
        sql, params = super().compile(node, select_format)
        # FIXME: Search lookup removed from Django-2.0, is this ok still?
        # # substitute MATCH() arguments with sphinx-escaped params
        # if isinstance(node, Search):
        #     search_text = sphinx_escape(params[0])
        #     sql = sql % search_text
        #     params = []

        return sql, params

    def get_order_by(self):
        res = super().get_order_by()

        order_by = []
        for expr, params in res:
            if isinstance(expr.expression, Random):
                # Replacing ORDER BY RAND() ASC to ORDER BY RAND()
                assert params[0] == 'RAND() ASC', "Expected ordering clause"
                params = ('RAND()',) + params[1:]
            order_by.append((expr, params))
        return order_by

    def get_group_by(self, select, order_by):
        res = super().get_group_by(select, order_by)

        # override GROUP BY columns for sphinxsearch's "GROUP N BY" support
        group_by = getattr(self.query, 'group_by', None)
        if group_by:
            # noinspection PyProtectedMember
            fields = self.query.model._meta.fields
            field_columns = [f.column for f in fields if f.attname in group_by]
            return [r for r in res if r[0] in field_columns]

        return res

    @staticmethod
    def _quote(s, negative=True):
        """ Adds quotes and negates to match lookup expressions."""
        prefix = '-' if negative else ''
        if s.startswith('"'):
            return s
        negative = s.startswith('-')
        if not negative:
            return '"%s"' % s
        s = s[1:]
        if s.startswith('"'):
            return '%s%s' % (prefix, s)
        return '%s"%s"' % (prefix, s)

    def _serialize(self, values_list):
        """ Serializes list of sphinx MATCH lookup expressions

        :param values_list: list of match lookup expressions
        :type values_list: str, list, tuple
        :return: MATCH expression
        :rtype: str
        """""
        if isinstance(values_list, str):
            return values_list

        def ensure_list(s):
            return [s] if isinstance(s, str) else s
        values_list = [item for s in values_list for item in ensure_list(s)]
        positive_list = filter(lambda s: not s.startswith('-'), values_list)
        negative_list = filter(lambda s: s.startswith('-'), values_list)

        positive = "|".join(map(self._quote, positive_list))
        if not positive_list:
            negative = '|'.join(self._quote(n, negative=False)
                                for n in negative_list)
            template = '%s -(%s)'
        else:
            negative = ' '.join(map(self._quote, negative_list))
            template = '%s %s'
        result = template % (positive.strip(' '), negative.strip(' '))
        return result.strip(' ')

    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        """ Patching final SQL query."""
        where, self.query.where = self.query.where, sqls.SphinxWhereNode()
        match = getattr(self.query, 'match', None)
        if match:
            # add match extra where
            self._add_match_extra(match)

        connection = self.connection

        where_sql, where_params = where.as_sql(self, connection)
        # moving where conditions to SELECT clause because of better support
        # of SQL expressions in sphinxsearch.

        if where_sql:
            # Without annotation queryset.count() receives 1 as where_result
            # and count it as aggregation result.
            self.query.add_annotation(
                sqls.SphinxWhereExpression(where_sql, where_params),
                '__where_result')
            # almost all where conditions are now in SELECT clause, so
            # WHERE should contain only test against that conditions are true
            self.query.add_extra(
                None, None,
                ['__where_result = %s'], (True,), None, None)

        sql, args = super().as_sql(with_limits, with_col_aliases)

        # empty SQL doesn't need patching
        if (sql, args) == ('', ()):
            return sql, args

        # removing unsupported by sphinxsearch OFFSET clause
        # replacing it with LIMIT <offset>, <limit>
        sql = re.sub(r'LIMIT (\d+) OFFSET (\d+)$', 'LIMIT \\2, \\1', sql)

        # patching GROUP BY clause
        group_limit = getattr(self.query, 'group_limit', '')
        group_by_ordering = self.get_group_ordering()
        if group_limit:
            # add GROUP <N> BY expression
            group_by = 'GROUP %s BY \\1' % group_limit
        else:
            group_by = 'GROUP BY \\1'
        if group_by_ordering:
            # add WITHIN GROUP ORDER BY expression
            group_by += group_by_ordering
        sql = re.sub(r'GROUP BY ((\w+)(, \w+)*)', group_by, sql)

        # adding sphinxsearch OPTION clause
        options = getattr(self.query, 'options', None)
        if options:
            keys = sorted(options.keys())
            values = [options[k] for k in keys if k not in self.safe_options]

            opts = []
            for k in keys:
                if k in self.safe_options:
                    opts.append("%s=%s" % (k, options[k]))
                else:
                    opts.append("%s=%%s" % k)
            sql += ' OPTION %s' % ', '.join(opts) or ''
            args += tuple(values)
        return sql, args

    def get_group_ordering(self):
        """ Returns group ordering clause.

        Formats WITHIN GROUP ORDER BY expression
        with columns in query.group_order_by
        """
        group_order_by = getattr(self.query, 'group_order_by', ())
        asc, desc = ORDER_DIR['ASC']
        if not group_order_by:
            return ''
        result = []
        for order_by in group_order_by:
            col, order = get_order_dir(order_by, asc)
            result.append("%s %s" % (col, order))
        return " WITHIN GROUP ORDER BY " + ", ".join(result)

    def _add_match_extra(self, match):
        """ adds MATCH clause to query.where """
        expression = []
        all_field_expr = []
        all_fields_lookup = match.get('*')

        # format expression to MATCH against any indexed fields
        if all_fields_lookup:
            if isinstance(all_fields_lookup, str):
                expression.append(all_fields_lookup)
                all_field_expr.append(all_fields_lookup)
            else:
                for value in all_fields_lookup:
                    value_str = self._serialize(value)
                    expression.append(value_str)
                    all_field_expr.append(value_str)

        # format expressions to MATCH against concrete fields
        for sphinx_attr, lookup in match.items():
            if sphinx_attr == '*':
                continue
            # noinspection PyProtectedMember
            field = self.query.model._meta.get_field(sphinx_attr)
            db_column = field.db_column or field.attname
            expression.append('@' + db_column)
            expression.append("(%s)" % self._serialize(lookup))

        # handle non-ascii characters in search expressions
        def decode(s):
            return s.decode("utf-8") if type(s) is bytes else s
        match_expr = u"MATCH('%s')" % u' '.join(map(decode, expression))

        # add MATCH() to query.where
        self.query.where.add(sqls.SphinxExtraWhere([match_expr], []), AND)


# Set SQLCompiler appropriately, so queries will use the correct compiler.
SQLCompiler = SphinxQLCompiler


class SQLInsertCompiler(compiler.SQLInsertCompiler, SphinxQLCompiler):
    pass


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SphinxQLCompiler):
    # noinspection PyMethodOverriding
    def as_sql(self):
        sql, params = super().as_sql()

        # empty SQL doesn't need patching
        if (sql, params) == ('', ()):
            return sql, params

        sql = re.sub(r'\(IN\((\w+),\s([\w\s%,]+)\)\)', '\\1 IN (\\2)', sql)
        return sql, params


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SphinxQLCompiler):

    # noinspection PyMethodOverriding
    def as_sql(self):
        node = self.is_single_row_update()
        # determine whether use UPDATE (only fixed-length fields) or
        # REPLACE (internal delete + insert) syntax
        need_replace = False
        if node:
            need_replace = self._has_string_fields()
        if node and need_replace:
            sql, args = self.as_replace(node)
        else:

            match = getattr(self.query, 'match', None)
            if match:
                # add match extra where
                self._add_match_extra(match)

            sql, args = super().as_sql()
        return sql, args

    def is_single_row_update(self):
        where = self.query.where
        match = getattr(self.query, 'match', {})
        node = None
        if len(where.children) == 1:
            node = where.children[0]
        elif match:
            # noinspection PyProtectedMember
            meta = self.query.model._meta
            pk_match = match.get(meta.pk.attname)
            if pk_match is not None:
                pk_value = list(pk_match.dict.keys())[0]
                return Exact(meta.pk.get_col(meta.db_table), pk_value)
        if not isinstance(node, Exact):
            node = None
        elif not node.lhs.field.primary_key:
            node = None
        return node

    def as_replace(self, where_node):
        """
        Performs single-row UPDATE as REPLACE INTO query.

        Must be used to change string attributes or indexed fields.
        """

        # It's a copy of compiler.SQLUpdateCompiler.as_sql method
        # that formats query more like INSERT than UPDATE
        self.pre_sql_setup()
        if not self.query.values:
            return '', ()
        table = self.query.base_table
        qn = self.quote_name_unless_alias
        result = ['REPLACE INTO %s' % qn(table)]
        # noinspection PyProtectedMember
        meta = self.query.model._meta
        self.query.values.append((meta.pk, self.query.model, where_node.rhs))
        columns, values, update_params = [], [], []

        for field, model, val in self.query.values:
            if hasattr(val, 'resolve_expression'):
                val = val.resolve_expression(self.query, allow_joins=False,
                                             for_save=True)
                if val.contains_aggregate:
                    raise FieldError(
                        "Aggregate functions are not allowed in this query")
            elif hasattr(val, 'prepare_database_save'):
                if field.rel:
                    val = field.get_db_prep_save(
                        val.prepare_database_save(field),
                        connection=self.connection,
                    )
                else:
                    raise TypeError(
                        "Database is trying to update a relational field "
                        "of type %s with a value of type %s. Make sure "
                        "you are setting the correct relations" %
                        (field.__class__.__name__, val.__class__.__name__))
            else:
                val = field.get_db_prep_save(val, connection=self.connection)

            # Getting the placeholder for the field.
            if hasattr(field, 'get_placeholder'):
                placeholder = field.get_placeholder(val, self, self.connection)
            else:
                placeholder = '%s'
            name = field.column
            columns.append(qn(name))
            if hasattr(val, 'as_sql'):
                sql, params = self.compile(val)
                values.append(sql)
                update_params.extend(params)
            elif val is not None:
                values.append(placeholder)
                update_params.append(val)
            else:
                values.append('NULL')
        if not values:
            return '', ()
        result.append('(')
        result.append(', '.join(columns))
        result.append(') VALUES (')
        result.append(', '.join(values))
        result.append(')')
        return ' '.join(result), tuple(update_params)

    def _has_string_fields(self):
        """ check whether query is updating text fields."""
        _excluded_update_fields = (
            models.CharField,
            models.TextField
        )
        for field, model, val in self.query.values:
            if isinstance(field, _excluded_update_fields):
                return True
        return False


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SphinxQLCompiler):
    pass



# coding: utf-8

from django.db.backends.mysql import base, creation
from django.db.backends.mysql.base import server_version_re
from django.utils.functional import cached_property


class SphinxOperations(base.DatabaseOperations):

    def regex_lookup(self, lookup_type):
        raise NotImplementedError()

    compiler_module = "sphinxsearch.backend.sphinx.compiler"

    def fulltext_search_sql(self, field_name):
        """ Formats full-text search expression."""
        return 'MATCH (\'@%s "%%s"\')' % field_name

    def force_no_ordering(self):
        """ Fix unsupported syntax "ORDER BY NULL"."""
        return []


class SphinxValidation(base.DatabaseValidation):
    def _check_sql_mode(self, **kwargs):
        """ Disable sql_mode validation because it's unsupported
        >>> import django.db
        >>> cursor = django.db.connection
        >>> cursor.execute("SELECT @@sql_mode")
        # Error here after parsing searchd response
        """
        return []


class SphinxCreation(creation.DatabaseCreation):

    def create_test_db(self, *args, **kwargs):
        # NOOP, test using regular sphinx database.
        if self.connection.settings_dict.get('TEST_NAME'):
            # initialize connection database name
            test_name = self.connection.settings_dict['TEST_NAME']
            self.connection.close()
            self.connection.settings_dict['NAME'] = test_name
            self.connection.cursor()
            return test_name
        return self.connection.settings_dict['NAME']

    def destroy_test_db(self, *args, **kwargs):
        # NOOP, we created nothing, nothing to destroy.
        return


class SphinxFeatures(base.DatabaseFeatures):
    # The following can be useful for unit testing, with multiple databases
    # configured in Django, if one of them does not support transactions,
    # Django will fall back to using clear/create
    # (instead of begin...rollback) between each test. The method Django
    # uses to detect transactions uses CREATE TABLE and DROP TABLE,
    # which ARE NOT supported by Sphinx, even though transactions ARE.
    # Therefore, we can just set this to True, and Django will use
    # transactions for clearing data between tests when all OTHER backends
    # support it.
    supports_transactions = True
    allows_group_by_pk = False
    uses_savepoints = False
    supports_column_check_constraints = False
    is_sql_auto_is_null_enabled = False


class DatabaseWrapper(base.DatabaseWrapper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ops = SphinxOperations(self)
        self.creation = SphinxCreation(self)
        self.features = SphinxFeatures(self)
        self.validation = SphinxValidation(self)

    def _start_transaction_under_autocommit(self):
        raise NotImplementedError()

    @cached_property
    def mysql_version(self):
        # Django>=1.10 makes if differently
        with self.temporary_connection():
            server_info = self.connection.get_server_info()
        match = server_version_re.match(server_info)
        if not match:
            raise Exception('Unable to determine MySQL version from version '
                            'string %r' % server_info)
        return tuple(int(x) for x in match.groups())



# coding: utf-8

# $Id: $


from django.conf import settings


class SphinxRouter(object):
    """
    Routes database operations for Sphinx model to the sphinx database connection.
    """

    def is_sphinx_model(self, model_or_obj):
        from sphinxsearch.models import SphinxModel
        from sphinxsearch.sql import SphinxModelBase
        if type(model_or_obj) is not SphinxModelBase:
            model = model_or_obj.__class__
        else:
            model = model_or_obj
        is_sphinx_model = issubclass(model, SphinxModel) or type(model) is SphinxModelBase
        return is_sphinx_model

    def db_for_read(self, model, **kwargs):
        if self.is_sphinx_model(model):
            return getattr(settings, 'SPHINX_DATABASE_NAME', 'sphinx')

    def db_for_write(self, model, **kwargs):
        if self.is_sphinx_model(model):
            return getattr(settings, 'SPHINX_DATABASE_NAME', 'sphinx')

    def allow_relation(self, obj1, obj2, **kwargs):
        # Allow all relations...
        return True



from distutils.core import setup

try:
    # noinspection PyPackageRequirements
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

setup(
    name='django_sphinxsearch',
    version='2.0.0',
    long_description=read_md('README.md'),
    packages=[
        'sphinxsearch',
        'sphinxsearch.backend',
        'sphinxsearch.backend.sphinx',
    ],
    url='http://github.com/rutube/django_sphinxsearch',
    license='Beerware',
    author='tumbler',
    author_email='zimbler@gmail.com',
    description='Sphinxsearch database backend for django>=2.0',
    setup_requires=[
        'Django>=2.0,<2.1',
        'mysqlclient>=1.3.3,<1.4.0'
    ],
)
